@@ -185,7 +185,7 @@ void fz_decouple_type3_font(fz_context *ctx, fz_font *font, void *t3doc)
 		return;
 
 	if (font->t3doc != t3doc)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "can't decouple type3 font from a different doc");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "can't decouple type3 font from a different doc");
 
 	font->t3doc = NULL;
 	free_resources(ctx, font);
@@ -211,9 +211,9 @@ fz_drop_font(fz_context *ctx, fz_font *font)
 
 	if (font->ft_face)
 	{
-		fz_lock(ctx, FZ_LOCK_FREETYPE);
+		fz_ft_lock(ctx);
 		fterr = FT_Done_Face((FT_Face)font->ft_face);
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
+		fz_ft_unlock(ctx);
 		if (fterr)
 			fz_warn(ctx, "FT_Done_Face(%s): %s", font->name, ft_error_string(fterr));
 		fz_drop_freetype(ctx);
@@ -316,7 +316,7 @@ struct fz_font_context
 	fz_font *base14[14];
 	fz_font *cjk[4];
 	struct { fz_font *serif, *sans; } fallback[256];
-	fz_font *symbol1, *symbol2, *math, *music;
+	fz_font *symbol1, *symbol2, *math, *music, *boxes;
 	fz_font *emoji;
 };
 
@@ -357,6 +357,33 @@ static void *ft_realloc(FT_Memory memory, long cur_size, long new_size, void *bl
 	return fz_realloc_no_throw(ctx, block, new_size);
 }
 
+void
+fz_ft_lock(fz_context *ctx)
+{
+	fz_lock(ctx, FZ_LOCK_FREETYPE);
+	fz_lock(ctx, FZ_LOCK_ALLOC);
+	assert(ctx->font->ftmemory.user == NULL);
+	ctx->font->ftmemory.user = ctx;
+	fz_unlock(ctx, FZ_LOCK_ALLOC);
+}
+
+void
+fz_ft_unlock(fz_context *ctx)
+{
+	fz_lock(ctx, FZ_LOCK_ALLOC);
+	ctx->font->ftmemory.user = NULL;
+	fz_unlock(ctx, FZ_LOCK_ALLOC);
+	fz_unlock(ctx, FZ_LOCK_FREETYPE);
+}
+
+int
+fz_ft_lock_held(fz_context *ctx)
+{
+	/* If this thread has locked the freetype lock already, then
+	 * the stored context will be this one. */
+	return (ctx->font->ftmemory.user == ctx);
+}
+
 void fz_new_font_context(fz_context *ctx)
 {
 	ctx->font = fz_malloc_struct(ctx, fz_font_context);
@@ -364,7 +391,7 @@ void fz_new_font_context(fz_context *ctx)
 	ctx->font->ftlib = NULL;
 	ctx->font->ftlib_refs = 0;
 	ctx->font->load_font = NULL;
-	ctx->font->ftmemory.user = ctx;
+	ctx->font->ftmemory.user = NULL;
 	ctx->font->ftmemory.alloc = ft_alloc;
 	ctx->font->ftmemory.free = ft_free;
 	ctx->font->ftmemory.realloc = ft_realloc;
@@ -401,6 +428,7 @@ void fz_drop_font_context(fz_context *ctx)
 		fz_drop_font(ctx, ctx->font->math);
 		fz_drop_font(ctx, ctx->font->music);
 		fz_drop_font(ctx, ctx->font->emoji);
+		fz_drop_font(ctx, ctx->font->boxes);
 		fz_free(ctx, ctx->font);
 		ctx->font = NULL;
 	}
@@ -428,6 +456,8 @@ fz_font *fz_load_system_font(fz_context *ctx, const char *name, int bold, int it
 		fz_catch(ctx)
 		{
 			fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+			fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+			fz_report_error(ctx);
 			font = NULL;
 		}
 	}
@@ -446,6 +476,8 @@ fz_font *fz_load_system_cjk_font(fz_context *ctx, const char *name, int ros, int
 		fz_catch(ctx)
 		{
 			fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+			fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+			fz_report_error(ctx);
 			font = NULL;
 		}
 	}
@@ -464,6 +496,8 @@ fz_font *fz_load_system_fallback_font(fz_context *ctx, int script, int language,
 		fz_catch(ctx)
 		{
 			fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+			fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+			fz_report_error(ctx);
 			font = NULL;
 		}
 	}
@@ -603,6 +637,19 @@ static fz_font *fz_load_fallback_emoji_font(fz_context *ctx)
 	return ctx->font->emoji;
 }
 
+static fz_font *fz_load_fallback_boxes_font(fz_context *ctx)
+{
+	const unsigned char *data;
+	int size;
+	if (!ctx->font->boxes)
+	{
+		data = fz_lookup_noto_boxes_font(ctx, &size);
+		if (data)
+			ctx->font->boxes = fz_new_font_from_memory(ctx, NULL, data, size, 0, 0);
+	}
+	return ctx->font->boxes;
+}
+
 static const struct ft_error ft_errors[] =
 {
 #include FT_ERRORS_H
@@ -626,11 +673,11 @@ fz_keep_freetype(fz_context *ctx)
 	int maj, min, pat;
 	fz_font_context *fct = ctx->font;
 
-	fz_lock(ctx, FZ_LOCK_FREETYPE);
+	fz_ft_lock(ctx);
 	if (fct->ftlib)
 	{
 		fct->ftlib_refs++;
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
+		fz_ft_unlock(ctx);
 		return;
 	}
 
@@ -638,8 +685,8 @@ fz_keep_freetype(fz_context *ctx)
 	if (fterr)
 	{
 		const char *mess = ft_error_string(fterr);
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot init freetype: %s", mess);
+		fz_ft_unlock(ctx);
+		fz_throw(ctx, FZ_ERROR_LIBRARY, "cannot init freetype: %s", mess);
 	}
 
 	FT_Add_Default_Modules(fct->ftlib);
@@ -650,12 +697,12 @@ fz_keep_freetype(fz_context *ctx)
 		fterr = FT_Done_Library(fct->ftlib);
 		if (fterr)
 			fz_warn(ctx, "FT_Done_Library(): %s", ft_error_string(fterr));
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
-		fz_throw(ctx, FZ_ERROR_GENERIC, "freetype version too old: %d.%d.%d", maj, min, pat);
+		fz_ft_unlock(ctx);
+		fz_throw(ctx, FZ_ERROR_LIBRARY, "freetype version too old: %d.%d.%d", maj, min, pat);
 	}
 
 	fct->ftlib_refs++;
-	fz_unlock(ctx, FZ_LOCK_FREETYPE);
+	fz_ft_unlock(ctx);
 }
 
 static void
@@ -664,7 +711,7 @@ fz_drop_freetype(fz_context *ctx)
 	int fterr;
 	fz_font_context *fct = ctx->font;
 
-	fz_lock(ctx, FZ_LOCK_FREETYPE);
+	fz_ft_lock(ctx);
 	if (--fct->ftlib_refs == 0)
 	{
 		fterr = FT_Done_Library(fct->ftlib);
@@ -672,7 +719,7 @@ fz_drop_freetype(fz_context *ctx)
 			fz_warn(ctx, "FT_Done_Library(): %s", ft_error_string(fterr));
 		fct->ftlib = NULL;
 	}
-	fz_unlock(ctx, FZ_LOCK_FREETYPE);
+	fz_ft_unlock(ctx);
 }
 
 fz_font *
@@ -688,13 +735,13 @@ fz_new_font_from_buffer(fz_context *ctx, const char *name, fz_buffer *buffer, in
 
 	fz_keep_freetype(ctx);
 
-	fz_lock(ctx, FZ_LOCK_FREETYPE);
+	fz_ft_lock(ctx);
 	fterr = FT_New_Memory_Face(ctx->font->ftlib, buffer->data, (FT_Long)buffer->len, index, &face);
-	fz_unlock(ctx, FZ_LOCK_FREETYPE);
+	fz_ft_unlock(ctx);
 	if (fterr)
 	{
 		fz_drop_freetype(ctx);
-		fz_throw(ctx, FZ_ERROR_GENERIC, "FT_New_Memory_Face(%s): %s", name, ft_error_string(fterr));
+		fz_throw(ctx, FZ_ERROR_LIBRARY, "FT_New_Memory_Face(%s): %s", name, ft_error_string(fterr));
 	}
 
 	if (!name)
@@ -724,9 +771,9 @@ fz_new_font_from_buffer(fz_context *ctx, const char *name, fz_buffer *buffer, in
 		font = fz_new_font(ctx, name, use_glyph_bbox, face->num_glyphs);
 	fz_catch(ctx)
 	{
-		fz_lock(ctx, FZ_LOCK_FREETYPE);
+		fz_ft_lock(ctx);
 		fterr = FT_Done_Face(face);
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
+		fz_ft_unlock(ctx);
 		if (fterr)
 			fz_warn(ctx, "FT_Done_Face(%s): %s", name, ft_error_string(fterr));
 		fz_drop_freetype(ctx);
@@ -751,6 +798,7 @@ fz_new_font_from_buffer(fz_context *ctx, const char *name, fz_buffer *buffer, in
 
 	if (FT_IS_SFNT(face))
 	{
+		fz_ft_lock(ctx);
 		os2 = FT_Get_Sfnt_Table(face, FT_SFNT_OS2);
 		if (os2)
 			font->flags.is_serif = !(os2->sFamilyClass & 2048); /* Class 8 is sans-serif */
@@ -770,6 +818,7 @@ fz_new_font_from_buffer(fz_context *ctx, const char *name, fz_buffer *buffer, in
 			if (tag == TTAG_GDEF || tag == TTAG_GPOS || tag == TTAG_GSUB)
 				font->flags.has_opentype = 1;
 		}
+		fz_ft_unlock(ctx);
 	}
 
 	if (name)
@@ -823,10 +872,17 @@ void fz_set_font_embedding(fz_context *ctx, fz_font *font, int embed)
 {
 	if (!font)
 		return;
-	if (font->flags.never_embed)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Embedding not permitted/possible");
-
-	font->flags.embed = !!embed;
+	if (embed)
+	{
+		if (font->flags.never_embed)
+			fz_warn(ctx, "not allowed to embed font: %s", font->name);
+		else
+			font->flags.embed = 1;
+	}
+	else
+	{
+		font->flags.embed = 0;
+	}
 }
 
 static int
@@ -871,7 +927,7 @@ fz_new_base14_font(fz_context *ctx, const char *name)
 			return fz_keep_font(ctx, ctx->font->base14[x]);
 		}
 	}
-	fz_throw(ctx, FZ_ERROR_GENERIC, "cannot find builtin font with name '%s'", name);
+	fz_throw(ctx, FZ_ERROR_ARGUMENT, "cannot find builtin font with name '%s'", name);
 }
 
 fz_font *
@@ -898,7 +954,7 @@ fz_new_cjk_font(fz_context *ctx, int ordering)
 			return fz_keep_font(ctx, ctx->font->cjk[ordering]);
 		}
 	}
-	fz_throw(ctx, FZ_ERROR_GENERIC, "cannot find builtin CJK font");
+	fz_throw(ctx, FZ_ERROR_ARGUMENT, "cannot find builtin CJK font");
 }
 
 fz_font *
@@ -909,7 +965,7 @@ fz_new_builtin_font(fz_context *ctx, const char *name, int is_bold, int is_itali
 	fz_font *font;
 	data = fz_lookup_builtin_font(ctx, name, is_bold, is_italic, &size);
 	if (!data)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "cannot find builtin font with name '%s'", name);
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "cannot find builtin font with name '%s'", name);
 	font = fz_new_font_from_memory(ctx, NULL, data, size, 0, 0);
 
 	/* Don't embed builtin fonts. */
@@ -929,9 +985,9 @@ fz_adjust_ft_glyph_width(fz_context *ctx, fz_font *font, int gid, fz_matrix *trm
 		float subw;
 		float realw;
 
-		fz_lock(ctx, FZ_LOCK_FREETYPE);
+		fz_ft_lock(ctx);
 		fterr = FT_Get_Advance(font->ft_face, gid, FT_LOAD_NO_SCALE | FT_LOAD_NO_HINTING | FT_LOAD_IGNORE_TRANSFORM, &adv);
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
+		fz_ft_unlock(ctx);
 		if (fterr && fterr != FT_Err_Invalid_Argument)
 			fz_warn(ctx, "FT_Get_Advance(%s,%d): %s", font->name, gid, ft_error_string(fterr));
 
@@ -985,7 +1041,7 @@ do_ft_render_glyph(fz_context *ctx, fz_font *font, int gid, fz_matrix trm, int a
 	if (font->flags.fake_italic)
 		trm = fz_pre_shear(trm, SHEAR, 0);
 
-	fz_lock(ctx, FZ_LOCK_FREETYPE);
+	fz_ft_lock(ctx);
 
 	if (aa == 0)
 	{
@@ -1070,7 +1126,7 @@ fz_render_ft_glyph_pixmap(fz_context *ctx, fz_font *font, int gid, fz_matrix trm
 
 	if (slot == NULL)
 	{
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
+		fz_ft_unlock(ctx);
 		return NULL;
 	}
 
@@ -1080,7 +1136,7 @@ fz_render_ft_glyph_pixmap(fz_context *ctx, fz_font *font, int gid, fz_matrix trm
 	}
 	fz_always(ctx)
 	{
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
+		fz_ft_unlock(ctx);
 	}
 	fz_catch(ctx)
 	{
@@ -1099,7 +1155,7 @@ fz_render_ft_glyph(fz_context *ctx, fz_font *font, int gid, fz_matrix trm, int a
 
 	if (slot == NULL)
 	{
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
+		fz_ft_unlock(ctx);
 		return NULL;
 	}
 
@@ -1109,7 +1165,7 @@ fz_render_ft_glyph(fz_context *ctx, fz_font *font, int gid, fz_matrix trm, int a
 	}
 	fz_always(ctx)
 	{
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
+		fz_ft_unlock(ctx);
 	}
 	fz_catch(ctx)
 	{
@@ -1146,7 +1202,7 @@ do_render_ft_stroked_glyph(fz_context *ctx, fz_font *font, int gid, fz_matrix tr
 	v.x = trm.e * 64;
 	v.y = trm.f * 64;
 
-	fz_lock(ctx, FZ_LOCK_FREETYPE);
+	fz_ft_lock(ctx);
 	fterr = FT_Set_Char_Size(face, 65536, 65536, 72, 72); /* should be 64, 64 */
 	if (fterr)
 	{
@@ -1222,7 +1278,7 @@ fz_render_ft_stroked_glyph(fz_context *ctx, fz_font *font, int gid, fz_matrix tr
 
 	if (bitmap == NULL)
 	{
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
+		fz_ft_unlock(ctx);
 		return NULL;
 	}
 
@@ -1233,7 +1289,7 @@ fz_render_ft_stroked_glyph(fz_context *ctx, fz_font *font, int gid, fz_matrix tr
 	fz_always(ctx)
 	{
 		FT_Done_Glyph(glyph);
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
+		fz_ft_unlock(ctx);
 	}
 	fz_catch(ctx)
 	{
@@ -1297,7 +1353,7 @@ fz_bound_ft_glyph(fz_context *ctx, fz_font *font, int gid)
 	v.x = trm.e * 65536;
 	v.y = trm.f * 65536;
 
-	fz_lock(ctx, FZ_LOCK_FREETYPE);
+	fz_ft_lock(ctx);
 	/* Set the char size to scale=face->units_per_EM to effectively give
 	 * us unscaled results. This avoids quantisation. We then apply the
 	 * scale ourselves below. */
@@ -1310,7 +1366,7 @@ fz_bound_ft_glyph(fz_context *ctx, fz_font *font, int gid)
 	if (fterr)
 	{
 		fz_warn(ctx, "FT_Load_Glyph(%s,%d,FT_LOAD_NO_HINTING): %s", font->name, gid, ft_error_string(fterr));
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
+		fz_ft_unlock(ctx);
 		bounds->x0 = bounds->x1 = trm.e;
 		bounds->y0 = bounds->y1 = trm.f;
 		return bounds;
@@ -1323,7 +1379,7 @@ fz_bound_ft_glyph(fz_context *ctx, fz_font *font, int gid)
 	}
 
 	FT_Outline_Get_CBox(&face->glyph->outline, &cbox);
-	fz_unlock(ctx, FZ_LOCK_FREETYPE);
+	fz_ft_unlock(ctx);
 	bounds->x0 = cbox.xMin * recip;
 	bounds->y0 = cbox.yMin * recip;
 	bounds->x1 = cbox.xMax * recip;
@@ -1419,7 +1475,7 @@ fz_outline_ft_glyph(fz_context *ctx, fz_font *font, int gid, fz_matrix trm)
 	if (font->flags.fake_italic)
 		trm = fz_pre_shear(trm, SHEAR, 0);
 
-	fz_lock(ctx, FZ_LOCK_FREETYPE);
+	fz_ft_lock(ctx);
 
 	fterr = FT_Set_Char_Size(face, scale, scale, 72, 72);
 	if (fterr)
@@ -1434,7 +1490,7 @@ fz_outline_ft_glyph(fz_context *ctx, fz_font *font, int gid, fz_matrix trm)
 	if (fterr)
 	{
 		fz_warn(ctx, "FT_Load_Glyph(%s,%d,FT_LOAD_IGNORE_TRANSFORM | FT_LOAD_NO_HINTING): %s", font->name, gid, ft_error_string(fterr));
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
+		fz_ft_unlock(ctx);
 		return NULL;
 	}
 
@@ -1456,7 +1512,7 @@ fz_outline_ft_glyph(fz_context *ctx, fz_font *font, int gid, fz_matrix trm)
 	}
 	fz_always(ctx)
 	{
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
+		fz_ft_unlock(ctx);
 	}
 	fz_catch(ctx)
 	{
@@ -1794,10 +1850,10 @@ fz_advance_ft_glyph_aux(fz_context *ctx, fz_font *font, int gid, int wmode, int
 	if (wmode)
 		mask |= FT_LOAD_VERTICAL_LAYOUT;
 	if (!locked)
-		fz_lock(ctx, FZ_LOCK_FREETYPE);
+		fz_ft_lock(ctx);
 	fterr = FT_Get_Advance(font->ft_face, gid, mask, &adv);
 	if (!locked)
-		fz_unlock(ctx, FZ_LOCK_FREETYPE);
+		fz_ft_unlock(ctx);
 	if (fterr && fterr != FT_Err_Invalid_Argument)
 	{
 		fz_warn(ctx, "FT_Get_Advance(%s,%d): %s", font->name, gid, ft_error_string(fterr));
@@ -1833,7 +1889,10 @@ fz_get_glyph_name(fz_context *ctx, fz_font *font, int glyph, char *buf, int size
 	{
 		if (FT_HAS_GLYPH_NAMES(face))
 		{
-			int fterr = FT_Get_Glyph_Name(face, glyph, buf, size);
+			int fterr;
+			fz_ft_lock(ctx);
+			fterr = FT_Get_Glyph_Name(face, glyph, buf, size);
+			fz_ft_unlock(ctx);
 			if (fterr)
 				fz_warn(ctx, "FT_Get_Glyph_Name(%s,%d): %s", font->name, glyph, ft_error_string(fterr));
 		}
@@ -1857,7 +1916,7 @@ fz_advance_glyph(fz_context *ctx, fz_font *font, int gid, int wmode)
 		{
 			float f;
 			int block = gid>>8;
-			fz_lock(ctx, FZ_LOCK_FREETYPE);
+			fz_ft_lock(ctx);
 			if (!font->advance_cache)
 			{
 				int n = (font->glyph_count+255)/256;
@@ -1865,7 +1924,7 @@ fz_advance_glyph(fz_context *ctx, fz_font *font, int gid, int wmode)
 					font->advance_cache = Memento_label(fz_malloc_array(ctx, n, float *), "font_advance_cache");
 				fz_catch(ctx)
 				{
-					fz_unlock(ctx, FZ_LOCK_FREETYPE);
+					fz_ft_unlock(ctx);
 					fz_rethrow(ctx);
 				}
 				memset(font->advance_cache, 0, n * sizeof(float *));
@@ -1877,7 +1936,7 @@ fz_advance_glyph(fz_context *ctx, fz_font *font, int gid, int wmode)
 					font->advance_cache[block] = Memento_label(fz_malloc_array(ctx, 256, float), "font_advance_cache");
 				fz_catch(ctx)
 				{
-					fz_unlock(ctx, FZ_LOCK_FREETYPE);
+					fz_ft_unlock(ctx);
 					fz_rethrow(ctx);
 				}
 				n = (block<<8)+256;
@@ -1888,7 +1947,7 @@ fz_advance_glyph(fz_context *ctx, fz_font *font, int gid, int wmode)
 					font->advance_cache[block][i] = fz_advance_ft_glyph_aux(ctx, font, (block<<8)+i, 0, 1);
 			}
 			f = font->advance_cache[block][gid & 255];
-			fz_unlock(ctx, FZ_LOCK_FREETYPE);
+			fz_ft_unlock(ctx);
 			return f;
 		}
 
@@ -1904,6 +1963,7 @@ fz_encode_character(fz_context *ctx, fz_font *font, int ucs)
 {
 	if (font->ft_face)
 	{
+		int idx;
 		if (ucs >= 0 && ucs < 0x10000)
 		{
 			int pg = ucs >> 8;
@@ -1912,12 +1972,17 @@ fz_encode_character(fz_context *ctx, fz_font *font, int ucs)
 			{
 				int i;
 				font->encoding_cache[pg] = fz_malloc_array(ctx, 256, uint16_t);
+				fz_ft_lock(ctx);
 				for (i = 0; i < 256; ++i)
 					font->encoding_cache[pg][i] = FT_Get_Char_Index(font->ft_face, (pg << 8) + i);
+				fz_ft_unlock(ctx);
 			}
 			return font->encoding_cache[pg][ix];
 		}
-		return FT_Get_Char_Index(font->ft_face, ucs);
+		fz_ft_lock(ctx);
+		idx = FT_Get_Char_Index(font->ft_face, ucs);
+		fz_ft_unlock(ctx);
+		return idx;
 	}
 	return ucs;
 }
@@ -1937,13 +2002,17 @@ fz_encode_character_sc(fz_context *ctx, fz_font *font, int unicode)
 			name = fz_glyph_name_from_unicode_sc(unicode);
 			if (name)
 			{
+				fz_ft_lock(ctx);
 				glyph = FT_Get_Name_Index(font->ft_face, (char*)name);
+				fz_ft_unlock(ctx);
 				if (glyph > 0)
 					return glyph;
 			}
 
 			sprintf(buf, "uni%04X.sc", unicode);
+			fz_ft_lock(ctx);
 			glyph = FT_Get_Name_Index(font->ft_face, buf);
+			fz_ft_unlock(ctx);
 			if (glyph > 0)
 				return glyph;
 		}
@@ -1957,9 +2026,11 @@ fz_encode_character_by_glyph_name(fz_context *ctx, fz_font *font, const char *gl
 	int glyph = 0;
 	if (font->ft_face)
 	{
+		fz_ft_lock(ctx);
 		glyph = ft_name_index(font->ft_face, glyphname);
 		if (glyph == 0)
 			glyph = ft_char_index(font->ft_face, fz_unicode_from_glyph_name(glyphname));
+		fz_ft_unlock(ctx);
 	}
 	// TODO: type3 fonts (not needed for now)
 	return glyph;
@@ -2075,6 +2146,14 @@ fz_encode_character_with_fallback(fz_context *ctx, fz_font *user_font, int unico
 			return *out_font = font, gid;
 	}
 
+	font = fz_load_fallback_boxes_font(ctx);
+	if (font)
+	{
+		gid = fz_encode_character(ctx, font, unicode);
+		if (gid > 0)
+			return *out_font = font, gid;
+	}
+
 	font = fz_new_base14_font(ctx, "Symbol");
 	if (font)
 	{
@@ -2140,7 +2219,7 @@ fz_shaper_data_t *fz_font_shaper_data(fz_context *ctx, fz_font *font)
 void fz_font_digest(fz_context *ctx, fz_font *font, unsigned char digest[16])
 {
 	if (!font->buffer)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "no font file for digest");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "no font file for digest");
 	if (!font->has_digest)
 	{
 		fz_md5_buffer(ctx, font->buffer, font->digest);
@@ -2174,7 +2253,7 @@ fz_extract_ttf_from_ttc(fz_context *ctx, fz_font *font)
 	uint32_t csumpos = 0;
 
 	if (!font || !font->buffer)
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Not a ttc");
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "missing input");
 
 	stream = fz_open_buffer(ctx, font->buffer);
 
@@ -2186,17 +2265,17 @@ fz_extract_ttf_from_ttc(fz_context *ctx, fz_font *font)
 	{
 		/* Signature */
 		if (fz_read_uint32(ctx, stream) != CHR('t','t','c','f'))
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Not a ttc");
+			fz_throw(ctx, FZ_ERROR_FORMAT, "Not a ttc");
 
 		/* Version */
 		tmp = fz_read_uint32(ctx, stream);
 		if (tmp != 0x10000 && tmp != 0x20000)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Unsupported TTC version");
+			fz_throw(ctx, FZ_ERROR_FORMAT, "Unsupported TTC version");
 
 		/* How many subfonts are there? */
 		tmp = fz_read_uint32(ctx, stream);
 		if ((uint32_t)font->subfont >= tmp || font->subfont < 0)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "Bad subfont in TTC");
+			fz_throw(ctx, FZ_ERROR_FORMAT, "Bad subfont in TTC");
 
 		/* Read through the index table until we get the one for our subfont. */
 		for (i = 0; i <= font->subfont; i++)
@@ -2289,3 +2368,21 @@ fz_extract_ttf_from_ttc(fz_context *ctx, fz_font *font)
 
 	return buf;
 }
+
+void fz_enumerate_font_cmap(fz_context *ctx, fz_font *font, fz_cmap_callback *cb, void *opaque)
+{
+	unsigned long ucs;
+	unsigned int gid;
+
+	if (font == NULL || font->ft_face == NULL)
+		return;
+
+	fz_ft_lock(ctx);
+	for (ucs = FT_Get_First_Char(font->ft_face, &gid); gid > 0; ucs = FT_Get_Next_Char(font->ft_face, ucs, &gid))
+	{
+		fz_ft_unlock(ctx);
+		cb(ctx, opaque, ucs, gid);
+		fz_ft_lock(ctx);
+	}
+	fz_ft_unlock(ctx);
+}
