@@ -1,9 +1,37 @@
-#include "mupdf/html.h"
+// Copyright (C) 2004-2022 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
+
+#include "mupdf/fitz.h"
+#include "html-imp.h"
+
+#include <string.h>
+
+#include "css-properties.h"
 
 struct lexbuf
 {
 	fz_context *ctx;
 	fz_pool *pool;
+	const unsigned char *start;
 	const unsigned char *s;
 	const char *file;
 	int line;
@@ -18,13 +46,93 @@ static fz_css_selector *parse_selector(struct lexbuf *buf);
 
 FZ_NORETURN static void fz_css_error(struct lexbuf *buf, const char *msg)
 {
-	fz_throw(buf->ctx, FZ_ERROR_SYNTAX, "css syntax error: %s (%s:%d)", msg, buf->file, buf->line);
+#define PRE_POST_SIZE 30
+	unsigned char text[PRE_POST_SIZE * 2 + 4];
+	unsigned char *d = text;
+	const unsigned char *s = buf->start;
+	int n;
+
+	/* We want to make a helpful fragment for the error message.
+	 * We want err_pos to be the point at which we just tripped
+	 * the error. err_pos needs to be at least 1 byte behind
+	 * our read pointer, as we've read that char. */
+	const unsigned char *err_pos = buf->s;
+	n = 1;
+
+	/* And if we're using lookahead, it's further behind. */
+	if (buf->lookahead >= CSS_KEYWORD)
+		n += buf->string_len;
+	else if (buf->lookahead != EOF)
+		n += 1;
+
+	/* But it can't be before the start of the buffer */
+	n = fz_mini(n, err_pos - buf->start);
+	err_pos -= n;
+
+	/* We're going to try to output:
+	 * <section prior to the error> ">" <the char that tripped> "<" <section after the error>
+	 */
+	/* Is the section prior to the error too long? If so, truncate it with an elipsis. */
+	n = sizeof(text)-1;
+	if (err_pos - s > n-PRE_POST_SIZE - 3)
+	{
+		*d++ = '.';
+		*d++ = '.';
+		*d++ = '.';
+		n -= 3;
+		s = err_pos - (n-PRE_POST_SIZE - 3);
+	}
+
+	/* Copy the prefix (if there is one) */
+	if (err_pos > s)
+	{
+		n = err_pos - s;
+		while (n)
+		{
+			unsigned char c = *s++;
+			*d++ = (c < 32 || c > 127) ? ' ' : c;
+			n--;
+		}
+	}
+
+	/* Marker, char, end marker */
+	*d++ = '>', n--;
+	if (*err_pos)
+		*d++ = *err_pos++, n--;
+	*d++ = '<', n--;
+
+	/* Postfix */
+	n = (int)strlen((const char *)err_pos);
+	if (n <= PRE_POST_SIZE)
+	{
+		while (n > 0)
+		{
+			unsigned char c = *err_pos++;
+			*d++ =  (c < 32 || c > 127) ? ' ' : c;
+			n--;
+		}
+	}
+	else
+	{
+		for (n = PRE_POST_SIZE-3; n > 0; n--)
+		{
+			unsigned char c = *err_pos++;
+			*d++ =  (c < 32 || c > 127) ? ' ' : c;
+		}
+
+		*d++ = '.';
+		*d++ = '.';
+		*d++ = '.';
+	}
+	*d = 0;
+
+	fz_throw(buf->ctx, FZ_ERROR_SYNTAX, "css syntax error: %s (%s:%d) (%s)", msg, buf->file, buf->line, text);
 }
 
 fz_css *fz_new_css(fz_context *ctx)
 {
 	fz_pool *pool = fz_new_pool(ctx);
-	fz_css *css;
+	fz_css *css = NULL;
 
 	fz_try(ctx)
 	{
@@ -80,13 +188,18 @@ static fz_css_condition *fz_new_css_condition(fz_context *ctx, fz_pool *pool, in
 
 static fz_css_property *fz_new_css_property(fz_context *ctx, fz_pool *pool, const char *name, fz_css_value *value, int spec)
 {
-	fz_css_property *prop = fz_pool_alloc(ctx, pool, sizeof *prop);
-	prop->name = fz_pool_strdup(ctx, pool, name);
-	prop->value = value;
-	prop->spec = spec;
-	prop->important = 0;
-	prop->next = NULL;
-	return prop;
+	struct css_property_info *info = css_property_lookup(name, strlen(name));
+	if (info)
+	{
+		fz_css_property *prop = fz_pool_alloc(ctx, pool, sizeof *prop);
+		prop->name = info->key;
+		prop->value = value;
+		prop->spec = spec;
+		prop->important = 0;
+		prop->next = NULL;
+		return prop;
+	}
+	return NULL;
 }
 
 static fz_css_value *fz_new_css_value_x(fz_context *ctx, fz_pool *pool, int type)
@@ -111,9 +224,12 @@ static fz_css_value *fz_new_css_value(fz_context *ctx, fz_pool *pool, int type,
 
 static void css_lex_next(struct lexbuf *buf)
 {
-	buf->c = *(buf->s++);
+	if (buf->c == 0)
+		return;
+	buf->s += fz_chartorune(&buf->c, (const char *)buf->s);
 	if (buf->c == '\n')
 		++buf->line;
+	buf->lookahead = EOF;
 }
 
 static void css_lex_init(fz_context *ctx, struct lexbuf *buf, fz_pool *pool, const char *s, const char *file)
@@ -121,7 +237,9 @@ static void css_lex_init(fz_context *ctx, struct lexbuf *buf, fz_pool *pool, con
 	buf->ctx = ctx;
 	buf->pool = pool;
 	buf->s = (const unsigned char *)s;
-	buf->c = 0;
+	buf->lookahead = EOF;
+	buf->start = buf->s;
+	buf->c = -1;
 	buf->file = file;
 	buf->line = 1;
 	css_lex_next(buf);
@@ -137,20 +255,23 @@ static inline int iswhite(int c)
 static int isnmstart(int c)
 {
 	return c == '\\' || c == '_' || (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') ||
-		(c >= 128 && c <= 255);
+		(c >= 128 && c <= UCS_MAX);
 }
 
 static int isnmchar(int c)
 {
 	return c == '\\' || c == '_' || (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') ||
-		(c >= '0' && c <= '9') || c == '-' || (c >= 128 && c <= 255);
+		(c >= '0' && c <= '9') || c == '-' || (c >= 128 && c <= UCS_MAX);
 }
 
 static void css_push_char(struct lexbuf *buf, int c)
 {
-	if (buf->string_len + 1 >= nelem(buf->string))
+	char out[4];
+	int n = fz_runetochar(out, c);
+	if (buf->string_len + n >= (int)nelem(buf->string))
 		fz_css_error(buf, "token too long");
-	buf->string[buf->string_len++] = c;
+	memcpy(buf->string + buf->string_len, out, n);
+	buf->string_len += n;
 }
 
 static int css_lex_accept(struct lexbuf *buf, int t)
@@ -354,8 +475,8 @@ restart:
 	{
 		if (css_lex_accept(buf, '-'))
 		{
-			css_lex_expect(buf, '>');
-			goto restart; /* ignore CDC */
+			if (css_lex_accept(buf, '>'))
+				goto restart; /* ignore CDC */
 		}
 		if (isnmstart(buf->c))
 		{
@@ -539,8 +660,11 @@ static fz_css_value *parse_expr(struct lexbuf *buf)
 		if (accept(buf, ','))
 		{
 			white(buf);
-			tail = tail->next = fz_new_css_value(buf->ctx, buf->pool, ',', ",");
-			tail = tail->next = parse_term(buf);
+			if (buf->lookahead != ';')
+			{
+				tail = tail->next = fz_new_css_value(buf->ctx, buf->pool, ',', ",");
+				tail = tail->next = parse_term(buf);
+			}
 		}
 		else if (accept(buf, '/'))
 		{
@@ -570,7 +694,10 @@ static fz_css_property *parse_declaration(struct lexbuf *buf)
 	expect(buf, ':');
 	white(buf);
 
-	p->value = parse_expr(buf);
+	if (p)
+		p->value = parse_expr(buf);
+	else
+		(void) parse_expr(buf);
 
 	/* !important */
 	if (accept(buf, '!'))
@@ -578,7 +705,8 @@ static fz_css_property *parse_declaration(struct lexbuf *buf)
 		white(buf);
 		if (buf->lookahead != CSS_KEYWORD || strcmp(buf->string, "important"))
 			fz_css_error(buf, "expected keyword 'important' after '!'");
-		p->important = 1;
+		if (p)
+			p->important = 1;
 		next(buf);
 		white(buf);
 	}
@@ -588,14 +716,17 @@ static fz_css_property *parse_declaration(struct lexbuf *buf)
 
 static fz_css_property *parse_declaration_list(struct lexbuf *buf)
 {
-	fz_css_property *head, *tail;
+	fz_css_property *head, *tail = NULL, *p;
 
 	white(buf);
 
 	if (buf->lookahead == '}' || buf->lookahead == EOF)
 		return NULL;
 
-	head = tail = parse_declaration(buf);
+	p = parse_declaration(buf);
+	if (p)
+		tail = p;
+	head = tail;
 
 	while (accept(buf, ';'))
 	{
@@ -603,7 +734,14 @@ static fz_css_property *parse_declaration_list(struct lexbuf *buf)
 
 		if (buf->lookahead != '}' && buf->lookahead != ';' && buf->lookahead != EOF)
 		{
-			tail = tail->next = parse_declaration(buf);
+			p = parse_declaration(buf);
+			if (p)
+			{
+				if (!head)
+					head = tail = p;
+				else
+					tail = tail->next = p;
+			}
 		}
 	}
 
@@ -631,7 +769,7 @@ static fz_css_condition *parse_condition(struct lexbuf *buf)
 
 	if (accept(buf, ':'))
 	{
-		accept(buf, ':'); /* swallow css3 :: syntax and pretend it's a normal pseudo-class */
+		(void)accept(buf, ':'); /* swallow css3 :: syntax and pretend it's a normal pseudo-class */
 		if (buf->lookahead != CSS_KEYWORD)
 			fz_css_error(buf, "expected keyword after ':'");
 		c = fz_new_css_condition(buf->ctx, buf->pool, ':', "pseudo", buf->string);
@@ -759,6 +897,7 @@ static fz_css_selector *parse_selector(struct lexbuf *buf)
 	{
 		if (accept(buf, ' '))
 		{
+			white(buf);
 			if (accept(buf, '+'))
 				sel = parse_combinator(buf, '+', sel);
 			else if (accept(buf, '>'))
@@ -806,8 +945,9 @@ static fz_css_rule *parse_ruleset(struct lexbuf *buf)
 	}
 	fz_catch(buf->ctx)
 	{
-		if (fz_caught(buf->ctx) != FZ_ERROR_SYNTAX)
-			fz_rethrow(buf->ctx);
+		fz_rethrow_unless(buf->ctx, FZ_ERROR_SYNTAX);
+		fz_report_error(buf->ctx);
+
 		while (buf->lookahead != EOF)
 		{
 			if (accept(buf, '}'))
@@ -943,6 +1083,16 @@ static fz_css_rule *parse_stylesheet(struct lexbuf *buf, fz_css_rule *chain)
 	return chain ? chain : tail;
 }
 
+const char *fz_css_property_name(int key)
+{
+	const char *name = "unknown";
+	size_t i;
+	for (i = 0; i < nelem(css_property_list); ++i)
+		if (*css_property_list[i].name && css_property_list[i].key == key)
+			name = css_property_list[i].name;
+	return name;
+}
+
 fz_css_property *fz_parse_css_properties(fz_context *ctx, fz_pool *pool, const char *source)
 {
 	struct lexbuf buf;
